/*============================================================================*/
/*                                                                            */
/*                                                                            */
/*                              reLOC 0.09-vegas                              */
/*                                                                            */
/*                   (C) Copyright 2011-2013 Pavel Surynek                    */
/*            http://www.surynek.com | <pavel.surynek@mff.cuni.cz>            */
/*                                                                            */
/*                                                                            */
/*============================================================================*/
// reloc.h / 0.09-vegas_099
/*----------------------------------------------------------------------------*/
//
// Relocation problem solving package - original development header.
//
/*----------------------------------------------------------------------------*/


#ifndef __RELOC_H__
#define __RELOC_H__

#include <vector>
#include <list>
#include <set>
#include <map>

#include "types.h"
#include "result.h"


using namespace std;
using namespace sReloc;


/*----------------------------------------------------------------------------*/

namespace sReloc
{

    class sEdge;
    class sArc;
    class sVertex;


/*----------------------------------------------------------------------------*/
// Global constants

    extern const sString sRELOC_INDENT;

    
/*----------------------------------------------------------------------------*/
// sVertex

    class sVertex
    {
    public:
	static const int ORDER_UNDEFINED = -1;

    public:
	typedef std::list<sArc*> Neighbors_list;

    public:
	sVertex();
	sVertex(int id);
	sVertex(const sVertex &vertex);
	const sVertex& operator=(const sVertex &vertex);

	int calc_NeighborCount(void) const;
	int calc_NeighborOrder(int vertex_id) const;
	int calc_NeighborID(int order) const;

    public:
	virtual void to_Screen(const sString &indent = "") const;
	virtual void to_Stream(FILE *fw, const sString &indent = "") const;

    public:
	int m_id;
	Neighbors_list m_Neighbors;
    };


/*----------------------------------------------------------------------------*/
// sArc

    class sArc
    {
    public:
	sArc();
	sArc(sEdge *edge, sVertex *target);
	sArc(const sArc &arc);
	const sArc& operator=(const sArc &arc);

    public:
	virtual void to_Screen(const sString &indent = "") const;
	virtual void to_Stream(FILE *fw, const sString &indent = "") const;

    public:
	sEdge *m_edge;
	sVertex *m_target;
    };


/*----------------------------------------------------------------------------*/
// sEdge

    class sEdge
    {
    public:
	sEdge(int id, sVertex *vertex_u, sVertex *vertex_v);
	sEdge(const sEdge &edge);
	const sEdge& operator=(const sEdge &edge);

    public:
	virtual void to_Screen(const sString &indent = "") const;
	virtual void to_Stream(FILE *fw, const sString &indent = "") const;

    public:
	int m_id;
	sArc m_arc_uv;
	sArc m_arc_vu;
    };


/*----------------------------------------------------------------------------*/
// sUndirectedGraph

    class sUndirectedGraph
    {
    public:
	typedef std::vector<int> Distances_vector; 
	typedef std::vector<Distances_vector> Distances_2d_vector;
	typedef std::vector<Distances_2d_vector> Distances_3d_vector;
	typedef std::vector<Distances_3d_vector> Distances_4d_vector;

	typedef std::vector<int> VertexIDs_vector;
	typedef std::vector<sVertex> Vertices_vector;
	typedef std::list<sEdge> Edges_list;
	typedef std::multimap<int, int, std::less<int> > VertexQueue_multimap;

	typedef std::pair<int, int> Vertex_pair;
	typedef std::multimap<int, Vertex_pair, std::less<int> > VertexPairQueue_multimap;

    public:
	sUndirectedGraph();
	sUndirectedGraph(int x_size, int y_size);
	sUndirectedGraph(int x_size, int y_size, double obstacle_prob);
	sUndirectedGraph(const sUndirectedGraph &undirected_graph);
	const sUndirectedGraph& operator=(const sUndirectedGraph &undirected_graph);
	/*----------------------------------------------------------------------------*/

	void add_Vertex(void);
	void add_Vertices(int Vertex_cnt = 1);

	int get_VertexCount(void) const;
	sVertex* get_Vertex(int id);
	const sVertex* get_Vertex(int id) const;

	void add_Edge(int u_id, int v_id);
	int get_EdgeCount(void) const;

	bool is_Adjacent(int u_id, int v_id) const;

	int calc_ShortestPath(int u_id, int v_id) const;
	void calc_SingleSourceShortestPaths(int s_id, Distances_vector &Distances) const;
	void calc_AllPairsShortestPaths(Distances_2d_vector &all_pairs_Distances) const;
	void calc_AllPairsShortestPaths(void);
	const Distances_2d_vector& get_AllPairsShortestPaths(void) const;

	int calc_ShortestCoopPath(int u1_id, int u2_id, int v1_id, int v2_id) const;
	void calc_SingleSourceShortestCoopPaths(int s1_id, int s2_id, Distances_2d_vector &Distances) const;
	void calc_AllPairsShortestCoopPaths(Distances_4d_vector &all_pairs_coop_Distances) const;
	void calc_AllPairsShortestCoopPaths(void);
	const Distances_4d_vector& get_AllPairsShortestCoopPaths(void) const;
	/*----------------------------------------------------------------------------*/

    public:
	virtual void to_Screen(const sString &indent = "") const;
	virtual void to_Screen_vertices(const sString &indent = "") const;

	virtual void to_Stream(FILE *fw, const sString &indent = "") const;
	virtual void to_Stream_vertices(FILE *fw, const sString &indent = "") const;

	virtual sResult to_File_multirobot(const sString &filename, const sString &indent = "") const;
	virtual void to_Stream_multirobot(FILE *fw, const sString &indent = "") const;

	virtual sResult from_File_multirobot(const sString &filename);
	virtual sResult from_Stream_multirobot(FILE *fr);
		
    public:
	int m_Edge_cnt;

	Vertices_vector m_Vertices;
	Edges_list m_Edges;

	bool m_all_pairs_distances_calculated;
	Distances_2d_vector m_all_pairs_Distances;

	bool m_all_pairs_coop_distances_calculated;
	Distances_4d_vector m_all_pairs_coop_Distances;
    }; 


/*----------------------------------------------------------------------------*/

} // namespace sReloc

#endif /* __RELOC_H__ */
