/*============================================================================*/
/*                                                                            */
/*                                                                            */
/*                              reLOC 0.09-vegas                              */
/*                                                                            */
/*                   (C) Copyright 2011-2013 Pavel Surynek                    */
/*            http://www.surynek.com | <pavel.surynek@mff.cuni.cz>            */
/*                                                                            */
/*                                                                            */
/*============================================================================*/
// solver_main.cpp / 0.09-vegas_099
/*----------------------------------------------------------------------------*/
//
// Solution generator - main program.
//
// Solves a given multirobot instance by the SAT solving technique from scratch.
// SATPlan iterative solution length increasing is adopted.
//
//
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/times.h>
#include <unistd.h>

#include "config.h"
#include "compile.h"
#include "defs.h"
#include "reloc.h"
#include "multirobot.h"
#include "compress.h"
#include "statistics.h"
#include "version.h"
#include "search.h"

#include "solver_main.h"


using namespace sReloc;


/*----------------------------------------------------------------------------*/

namespace sReloc
{


/*----------------------------------------------------------------------------*/

  sCommandParameters::sCommandParameters()
      : m_base_strategy(sCommandParameters::STRATEGY_LINEAR_UP)
      , m_cnf_encoding(sMultirobotSolutionCompressor::ENCODING_UNDEFINED)
      , m_makespan_upper_bound(8)
      , m_minisat_timeout(-1)
      , m_total_timeout(600)
  {
      // nothing
  }


/*----------------------------------------------------------------------------*/

    void print_IntroductoryMessage(void)
    {
	printf("================================================================\n");
	printf("%s : Multirobot Solution solver\n", sPRODUCT);
	printf("%s\n", sCOPYRIGHT);
	printf("----------------------------------------------------------------\n");
    }


    void print_ConcludingMessage(void)
    {
	printf("----------------------------------------------------------------\n");
    }


    void print_Help(void)
    {
	printf("Usage:\n");
	printf("solver_reLOC  --input-file=<string>\n");
	printf("             [--output-file=<string>]\n");
	printf("             [--total-timeout=<int>]\n");
	printf("             [--minisat-timeout=<int>]\n");
	printf("             [--pddl-domain-file=<string>]\n");
	printf("             [--pddl-problem-file=<string>]\n");
	printf("             [--encoding={inverse|advanced|differential|bijection|Hadvanced|Hdifferential|Hbijection|\n");
	printf("                          bitwise|flow|matching|Hmatching|direct|Hdirect|simplicial|Hsimplicial|decomposed|independent}]\n");
	printf("             [--strategy={linear-down|linear-up|binary}]\n");
	printf("             [--makespan-limit=<int>]\n");
	printf("\n");
	printf("Examples:\n");
	printf("solver_reLOC --input-file=grid_02.txt\n");
	printf("             --output-file=grid_02.out\n");
	printf("\n");
	printf("Defaults: --minisat-timeout=-1 (unlimited)\n");
	printf("          --base-strategy=linear-up\n");
	printf("          --encoding=inverse\n");	
	printf("          --total-timeout=600\n");
	printf("          --makespan-limit=8\n");
    }


    sResult solve_MultirobotInstance_SAT(const sCommandParameters &command_parameters)
    {
	sUndirectedGraph environment;

	printf("Reading graph...\n");
	sResult result = environment.from_File_multirobot(command_parameters.m_input_filename);
	if (sFAILED(result))
	{
	    printf("Error: Reading graph from file %s failed (code = %d).\n", command_parameters.m_input_filename.c_str(), result);
	    return result;
	}

	printf("Reading initial arrangement...\n");
	sRobotArrangement initial_arrangement;
	result = initial_arrangement.from_File_multirobot(command_parameters.m_input_filename);
	if (sFAILED(result))
	{
	    printf("Error: Reading arrangement from file %s failed (code = %d).\n", command_parameters.m_input_filename.c_str(), result);
	    return result;
	}

	printf("Reading goal arrangement...\n");
	sRobotArrangement goal_arrangement;
	sRobotGoal robot_goal;
	/*
	result = goal_arrangement.from_File_multirobot(command_parameters.m_input_filename, 2);
	if (sFAILED(result))
	{
	    printf("Error: Reading arrangement from file %s failed (code = %d).\n", command_parameters.m_input_filename.c_str(), result);
	    return result;
	}
	*/
	result = robot_goal.from_File_multirobot(command_parameters.m_input_filename, 2);
	if (sFAILED(result))
	{
	    printf("Error: Reading arrangement from file %s failed (code = %d).\n", command_parameters.m_input_filename.c_str(), result);
	    return result;
	}


	/*
	original_solution.execute_Solution(initial_arrangement, goal_arrangement);
	*/

	//	sMultirobotInstance instance(environment, initial_arrangement, goal_arrangement);
	sMultirobotInstance instance(environment, initial_arrangement, robot_goal);

	if (!command_parameters.m_pddl_problem_filename.empty())
	{
	    result = instance.to_File_problemPDDL(command_parameters.m_pddl_problem_filename);

	    if (sFAILED(result))
	    {
		printf("Error: Failed to write PDDL problem file %s (code = %d).\n", command_parameters.m_pddl_problem_filename.c_str(), result);
		return result;
	    }
	}
	if (!command_parameters.m_pddl_domain_filename.empty())
	{
	    result = instance.to_File_domainPDDL(command_parameters.m_pddl_domain_filename);

	    if (sFAILED(result))
	    {
		printf("Error: Failed to write PDDL domain file %s (code = %d).\n", command_parameters.m_pddl_domain_filename.c_str(), result);
		return result;
	    }
	}

	#ifdef sVERBOSE
	{
	    environment.to_Screen();
	    initial_arrangement.to_Screen();
	    goal_arrangement.to_Screen();
	    robot_goal.to_Screen();
	}
	#endif


	if (!command_parameters.m_output_filename.empty())
	{
	    int optimal_makespan;
	    sMultirobotSolution optimal_solution;

	    if (command_parameters.m_cnf_encoding != sMultirobotSolutionCompressor::ENCODING_UNDEFINED)
	    {
		if (command_parameters.m_cnf_encoding != sMultirobotSolutionCompressor::ENCODING_UNUSED)
		{
		    sMultirobotSolutionCompressor compressor("../../sat/glucose3"/*"../../sat/minisat_static" , "../../sat/glucose_static"*/,
							     command_parameters.m_minisat_timeout,
							     command_parameters.m_total_timeout,
							     command_parameters.m_makespan_upper_bound,
							     sDEFAULT_N_PARALLEL_THREADS,
							     command_parameters.m_cnf_encoding);		    
		    switch (command_parameters.m_base_strategy)
		    {
		    case sCommandParameters::STRATEGY_LINEAR_DOWN:
		    {
			result = compressor.compute_OptimalSolution(initial_arrangement,
								    robot_goal,
								    environment,
								    command_parameters.m_makespan_upper_bound,
								    optimal_makespan,
								    optimal_solution);
			break;
		    }
		    case sCommandParameters::STRATEGY_LINEAR_UP:
		    {
			result = compressor.compute_OptimalSolution_(initial_arrangement,
								     robot_goal,
								     environment,
								     command_parameters.m_makespan_upper_bound,
								     optimal_makespan,
								     optimal_solution);
			break;
		    }
		    case sCommandParameters::STRATEGY_BINARY:
		    {
			result = compressor.compute_OptimalSolution(initial_arrangement,
								    robot_goal,
								    environment,
								    2,
								    command_parameters.m_makespan_upper_bound,
								    optimal_makespan,
								    optimal_solution);
			
			break;
		    }
		    default:
		    {
			sASSERT(false);
		    }
		    }
		    
		    if (sFAILED(result))
		    {
			printf("Error: Solution optimization failed (code = %d).\n", result);
			return result;
		    }
		    printf("Computed optimal makespan:%d\n", optimal_makespan);
		    printf("Makespan optimal solution:\n");
		    optimal_solution.to_Screen();
		    optimal_solution.to_File_multirobot(command_parameters.m_output_filename.c_str());
		}
		else
		{
		    sMultirobotSolver_DecomposedAstar solver_Astar(command_parameters.m_total_timeout);
		    sAstarHeuristic_Distance distance_heuristic;

		    solver_Astar.setup_Solver(&distance_heuristic);
		    solver_Astar.setup_Instance(sMultirobotInstance(environment, initial_arrangement, robot_goal));
	       
		    switch (solver_Astar.solve_Instance(optimal_solution))
		    {
		    case sMultirobotSolver_DecomposedAstar::RESULT_SOLVABLE:
		    {
			optimal_makespan = optimal_solution.get_StepCount();
			
			printf("Computed optimal makespan:%d\n", optimal_makespan);
			printf("Makespan optimal solution:\n");
			optimal_solution.to_Screen();
			optimal_solution.to_File_multirobot(command_parameters.m_output_filename.c_str());
			break;
		    }
		    case sMultirobotSolver_DecomposedAstar::RESULT_UNSOLVABLE:
		    {
			printf("Unable to find solution.\n");
			break;
		    }
		    case sMultirobotSolver_DecomposedAstar::RESULT_INDETERMINATE:
		    {
			printf("Cannot decide existence of solution..\n");
			break;
		    }
		    default:
		    {
			sASSERT(false);
		    }
		    }
		}
	    }
	    else
	    {
		sMultirobotSolver_DecomposedAstar solver_Astar(command_parameters.m_total_timeout);
		sAstarHeuristic_Distance distance_heuristic;

		solver_Astar.setup_Solver(&distance_heuristic);
		solver_Astar.setup_Instance(sMultirobotInstance(environment, initial_arrangement, robot_goal));
	       
		switch (solver_Astar.solve_DecomposedInstance(optimal_solution))
		{
		case sMultirobotSolver_DecomposedAstar::RESULT_SOLVABLE:
		{
		    optimal_makespan = optimal_solution.get_StepCount();

		    printf("Computed optimal makespan:%d\n", optimal_makespan);
		    printf("Makespan optimal solution:\n");
		    optimal_solution.to_Screen();
		    optimal_solution.to_File_multirobot(command_parameters.m_output_filename.c_str());
		    break;
		}
		case sMultirobotSolver_DecomposedAstar::RESULT_UNSOLVABLE:
		{
		    printf("Unable to find solution.\n");
		    break;
		}
		case sMultirobotSolver_DecomposedAstar::RESULT_INDETERMINATE:
		{
		    printf("Cannot decide existence of solution..\n");
		    break;
		}
		default:
		{
		    sASSERT(false);
		}
		}
	    }
	    sMultirobotSolutionAnalyzer solution_analyzer;
	    solution_analyzer.analyze_Solution(optimal_solution, initial_arrangement, environment);
	    solution_analyzer.to_Screen();
	}
	s_GlobalPhaseStatistics.to_Screen();

	return sRESULT_SUCCESS;
    }


    sResult parse_CommandLineParameter(const sString &parameter, sCommandParameters &command_parameters)
    {
	if (parameter.find("--input-file=") == 0)
	{
	    command_parameters.m_input_filename = parameter.substr(13, parameter.size());
	}
	else if (parameter.find("--output-file=") == 0)
	{
	    command_parameters.m_output_filename = parameter.substr(14, parameter.size());
	}
	else if (parameter.find("--pddl-domain-file=") == 0)
	{
	    command_parameters.m_pddl_domain_filename = parameter.substr(19, parameter.size());
	}
	else if (parameter.find("--pddl-problem-file=") == 0)
	{
	    command_parameters.m_pddl_problem_filename = parameter.substr(20, parameter.size());
	}
	else if (parameter.find("--makespan-limit=") == 0)
	{
	    command_parameters.m_makespan_upper_bound = sInt_32_from_String(parameter.substr(17, parameter.size()));
	}
	else if (parameter.find("--encoding=") == 0)
	{
	    sString encoding_str = parameter.substr(11, parameter.size());
	    if (encoding_str == "inverse")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_INVERSE;
	    }
	    else if (encoding_str == "advanced")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_ADVANCED;
	    }
	    else if (encoding_str == "differential")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_DIFFERENTIAL;
	    }
	    else if (encoding_str == "bijection")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_BIJECTION;
	    }
	    else if (encoding_str == "Hadvanced")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_HEURISTIC_ADVANCED;
	    }
	    else if (encoding_str == "Hdifferential")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_HEURISTIC_DIFFERENTIAL;
	    }
	    else if (encoding_str == "Hbijection")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_HEURISTIC_BIJECTION;
	    }
	    else if (encoding_str == "bitwise")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_BITWISE;
	    }
	    else if (encoding_str == "flow")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_FLOW;
	    }
	    else if (encoding_str == "matching")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_MATCHING;
	    }
	    else if (encoding_str == "Hmatching")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_HEURISTIC_MATCHING;
	    }
	    else if (encoding_str == "direct")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_DIRECT;
	    }
	    else if (encoding_str == "Hdirect")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_HEURISTIC_DIRECT;
	    }
	    else if (encoding_str == "simplicial")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_SIMPLICIAL;
	    }
	    else if (encoding_str == "Hsimplicial")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_HEURISTIC_SIMPLICIAL;
	    }
	    else if (encoding_str == "decomposed")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_UNUSED;
	    }
	    else if (encoding_str == "independent")
	    {
		command_parameters.m_cnf_encoding = sMultirobotSolutionCompressor::ENCODING_UNDEFINED;
	    }
	    else
	    {
		return sOPTIMIZER_PROGRAM_UNRECOGNIZED_PARAMETER_ERROR;
	    }
	}
	else if (parameter.find("--strategy=") == 0)
	{
	    sString encoding_str = parameter.substr(11, parameter.size());

	    if (encoding_str == "linear-down")
	    {
		command_parameters.m_base_strategy = sCommandParameters::STRATEGY_LINEAR_DOWN;
	    }
	    else if (encoding_str == "linear-up")
	    {
		command_parameters.m_base_strategy = sCommandParameters::STRATEGY_LINEAR_UP;
	    }
	    else if (encoding_str == "binary")
	    {
		command_parameters.m_base_strategy = sCommandParameters::STRATEGY_BINARY;
	    }
	    else
	    {
		return sOPTIMIZER_PROGRAM_UNRECOGNIZED_PARAMETER_ERROR;
	    }
	}
	else if (parameter.find("--minisat-timeout=") == 0)
	{
	    command_parameters.m_minisat_timeout = sInt_32_from_String(parameter.substr(18, parameter.size()));
	}
	else if (parameter.find("--total-timeout=") == 0)
	{
	    command_parameters.m_total_timeout = sInt_32_from_String(parameter.substr(16, parameter.size()));
	}
	else
	{
	    return sSOLVER_PROGRAM_UNRECOGNIZED_PARAMETER_ERROR;
	}
	return sRESULT_SUCCESS;
    }


/*----------------------------------------------------------------------------*/

} // namespace sReloc


/*----------------------------------------------------------------------------*/
// main program

int main(int argc, char **argv)
{
    sResult result;
    sCommandParameters command_parameters;

    print_IntroductoryMessage();

    if (argc >= 2 && argc <= 8)
    {
	for (int i = 1; i < argc; ++i)
	{
	    result = parse_CommandLineParameter(argv[i], command_parameters);
	    if (sFAILED(result))
	    {
		printf("Error: Cannot parse command line parameters (code = %d).\n", result);
		print_Help();
		return result;
	    }
	}
	result = solve_MultirobotInstance_SAT(command_parameters);

	if (sFAILED(result))
	{
	    return result;
	}
    }
    else
    {
	print_Help();
    }
    print_ConcludingMessage();

    return sRESULT_SUCCESS;
}

